from typing import Any
from typing import TypeVar

# Sentinel object used to check if the __config_class__ attribute has been set on the class
SENTINEL = object()

# Generic type variable T
T = TypeVar("T")


def mark(cls: type[T]) -> type[T]:
    """
    Mark a given class as a config class.

    This function sets the attribute __config_class__ on the class,
    with the value SENTINEL. If the attribute __config_class__ already exists,
    but is not the SENTINEL, it raises a TypeError indicating that
    the attribute has an unexpected value.

    :param cls: The class to be marked as a config class.
    :type cls: type[T]
    :return: The class, marked as a config class.
    :rtype: type[T]
    """
    if getattr(cls, "__config_class__", SENTINEL) is not SENTINEL:
        raise TypeError("Variable __config_class__ has an unexpected value")
    setattr(cls, "__config_class__", SENTINEL)
    return cls


def is_config(obj: Any) -> bool:
    """
    Check if an object is marked as a config class.

    This function checks if the attribute __config_class__
    exists on the object and if it is equal to SENTINEL.

    :param obj: The object to be checked.
    :return: True if the object is marked as a config class, False otherwise.
    :rtype: bool
    """
    return getattr(obj, "__config_class__", None) is SENTINEL



import typing

import sqlalchemy as sa
from sqlalchemy.ext.asyncio import create_async_engine

from gyver.config import from_config
from gyver.database.typedef import Driver

from .config import DatabaseConfig
from .config import DefaultDatabaseConfig
from .context import AsyncSaAdapter
from .context import SaAdapter
from .utils import make_uri


class SyncDatabaseProvider(SaAdapter):
    def __init__(self, config: typing.Optional[DatabaseConfig] = None) -> None:
        self._config = config or from_config(DefaultDatabaseConfig)
        db_kwargs = (
            {}
            if self._config.driver is Driver.SQLITE
            else {
                "pool_size": self._config.pool_size,
                "pool_recycle": self._config.pool_recycle,
                "max_overflow": self._config.max_overflow,
            }
        )

        super().__init__(
            engine=sa.create_engine(make_uri(self.config, sync=True), **db_kwargs)
        )

    @property
    def config(self):
        return self._config


class AsyncDatabaseProvider(AsyncSaAdapter):
    def __init__(self, config: typing.Optional[DatabaseConfig] = None) -> None:
        self._config = config or from_config(DefaultDatabaseConfig)
        db_kwargs = (
            {}
            if self._config.driver is Driver.SQLITE
            else {
                "pool_size": self._config.pool_size,
                "pool_recycle": self._config.pool_recycle,
                "max_overflow": self._config.max_overflow,
            }
        )

        super().__init__(engine=create_async_engine(make_uri(self.config), **db_kwargs))

    @property
    def config(self):
        return self._config



import typing

from starlette.applications import Starlette
from starlette.requests import Request

from gyver.database import AsyncSaContext
from gyver.database import SaContext
from gyver.database.config import DatabaseConfig
from gyver.database.provider import AsyncDatabaseProvider
from gyver.database.provider import SyncDatabaseProvider

TrxOptions = typing.Optional[typing.Literal["open", "begin"]]


def setup_database_session(
    app: Starlette,
    config: typing.Optional[DatabaseConfig] = None,
    sync: bool = False,
):
    provider_class = SyncDatabaseProvider if sync else AsyncDatabaseProvider

    def _setup():
        provider = provider_class(config)
        app.state.gyver_database = provider

    app.add_event_handler("startup", _setup)


def get_async_context(transaction_on: TrxOptions = "open"):
    def _get(request: Request) -> AsyncSaContext:
        provider = typing.cast(AsyncDatabaseProvider, request.app.state.gyver_database)
        return provider.context(transaction_on=transaction_on)

    return _get


def get_sync_context(transaction_on: TrxOptions = "open"):
    def _get(request: Request) -> SaContext:
        provider = typing.cast(SyncDatabaseProvider, request.app.state.gyver_database)
        return provider.context(transaction_on=transaction_on)

    return _get



from pathlib import Path
from typing import Any
from typing import Optional
from typing import Sequence
from typing import TypeVar

from pydantic.fields import ModelField

from gyver.config.adapter.factory import AdapterConfigFactory
from gyver.config.adapter.mark import mark
from gyver.config.adapter.pydantic import PydanticResolverStrategy
from gyver.model import Model
from gyver.utils import finder
from gyver.utils.helpers import DeprecatedClass
from gyver.utils.helpers import deprecated

from .config import Config


@mark
class ProviderConfig(Model):
    __prefix__ = ""
    __without_prefix__ = ()

    class Config:
        alias_generator = str.upper
        fields = {"exclude": {}}


ProviderT = TypeVar("ProviderT", bound=ProviderConfig)

_default_config = Config()


@deprecated
def from_config(
    provider: type[ProviderT],
    *,
    __config__: Config = _default_config,
    **presets: Any,
) -> ProviderT:
    return ConfigLoader(__config__).load(provider, **presets)


class ConfigLoader(DeprecatedClass):
    def __init__(
        self,
        config: Config = _default_config,
        prefix: Optional[str] = None,
        without_prefix: Sequence[str] = (),
    ):
        super().__init__()
        self._adapter_factory = AdapterConfigFactory(config)
        self._prefix = prefix
        self._without_prefix = without_prefix

    def load(self, model_cls: type[ProviderT], **presets: Any) -> ProviderT:
        return self._adapter_factory.load(
            model_cls, __prefix__=self._prefix or "", presets=presets
        )

    def resolve_names(
        self, model_cls: type[ProviderConfig], field: ModelField
    ) -> Sequence[str]:
        return self._adapter_factory.resolve_names(
            model_cls, PydanticResolverStrategy(field), self._prefix or ""
        )

    @classmethod
    def resolve_confignames(
        cls,
        root: Path,
    ) -> dict[type[ProviderConfig], tuple[Sequence[str], ...]]:
        """
        The resolve_confignames function resolves the names of all environment
        variables required by the configs in
        a ProviderConfig subclass.
        It returns them in a dict of {class: (configs)}.

        :param cls: Call the class_validator function, which is used to validate that a
        class has all of the required attributes
        :param root: Path: Specify the root directory of the project
        :return: A dictionary of {class: (configs)}
        """

        validator = finder.class_validator(ProviderConfig)
        provider_finder = finder.Finder(validator, root)
        provider_finder.find()
        tempself = cls()
        return {
            provider: tuple(
                tempself.resolve_names(provider, field)
                for field in provider.__fields__.values()
            )
            for provider in provider_finder.output.values()
        }



from .adapter import AdapterConfigFactory
from .adapter.mark import mark
from .config import MISSING
from .config import Config
from .config import EnvMapping
from .envconfig import DotFile
from .envconfig import EnvConfig
from .provider import ConfigLoader
from .provider import ProviderConfig
from .provider import from_config
from .typedef import Env
from .utils import boolean_cast

__all__ = [
    "boolean_cast",
    "Config",
    "EnvMapping",
    "MISSING",
    "ProviderConfig",
    "AdapterConfigFactory",
    "mark",
    "ConfigLoader",
    "from_config",
    "EnvConfig",
    "Env",
    "DotFile",
]



from contextlib import suppress
from dataclasses import is_dataclass
from pathlib import Path
from typing import Any
from typing import Mapping
from typing import Optional
from typing import Sequence
from typing import TypeVar
from typing import get_args
from typing import get_origin

from pydantic import BaseModel

from gyver.config.config import MISSING
from gyver.config.config import Config
from gyver.config.utils import boolean_cast
from gyver.exc import MissingName
from gyver.utils import finder
from gyver.utils import json
from gyver.utils import panic
from gyver.utils.strings import make_lex_separator

from .attrs import AttrsResolverStrategy
from .dataclass import DataclassResolverStrategy
from .interface import FieldResolverStrategy
from .mark import is_config
from .pydantic import PydanticResolverStrategy

_default_config = Config()

T = TypeVar("T")


def _try_each(*names: str, default: Any, cast: Any, config: Config):
    for name in names:
        with suppress(MissingName):
            return config(name, cast)
    if default is not MISSING:
        return default
    raise panic(MissingName, f"{', '.join(names)} not found and no default was given")


def _resolve_cast(outer_type: type):
    _sequences = (list, tuple, set)
    origin = get_origin(outer_type)
    if outer_type is bool:
        return boolean_cast
    if origin is None:
        return (
            make_lex_separator(outer_type) if outer_type in _sequences else outer_type
        )
    if (origin := get_origin(outer_type)) in _sequences:
        args = get_args(outer_type)
        cast = args[0] if args else str
        return make_lex_separator(origin, cast)  # type: ignore
    return _loads if dict in (origin, outer_type) else origin


def _loads(val: Any) -> Any:
    return json.loads(val) if isinstance(val, str) else val


class AdapterConfigFactory:
    def __init__(self, config: Config = _default_config) -> None:
        self._config = config

    def get_strategy_class(self, config_class: type) -> type[FieldResolverStrategy]:
        if is_dataclass(config_class):
            return DataclassResolverStrategy
        elif issubclass(config_class, BaseModel):
            return PydanticResolverStrategy
        elif hasattr(config_class, "__attrs_attrs__"):
            return AttrsResolverStrategy
        raise ValueError("Unknown class definition")

    def load(
        self,
        model_cls: type[T],
        __prefix__: str = "",
        *,
        presets: Optional[Mapping[str, Any]] = None,
        **defaults: Any,
    ) -> T:
        presets = presets or {}
        strategy_class = self.get_strategy_class(model_cls)
        resolvers = tuple(
            resolver
            for field in strategy_class.iterfield(model_cls)
            if (resolver := strategy_class(field)).init_name() not in presets
        )
        result = {
            resolver.init_name(): self._get_value(
                model_cls, resolver, __prefix__, defaults
            )
            for resolver in resolvers
        }
        return model_cls(**result | presets)

    def _get_value(
        self,
        model_cls: type,
        resolver: FieldResolverStrategy,
        prefix: str,
        defaults: Mapping[str, Any],
    ):
        names = self.resolve_names(model_cls, resolver, prefix)
        default = next(
            (result for name in names if (result := defaults.get(name))),
            resolver.default(),
        )
        cast = _resolve_cast(resolver.cast())
        return _try_each(*names, default=default, cast=cast, config=self._config)

    def resolve_names(
        self, model_cls: type, resolver: FieldResolverStrategy, prefix: str
    ) -> Sequence[str]:
        names = resolver.names()
        prefix = prefix or getattr(model_cls, "__prefix__", "")
        prefix = prefix.removesuffix("_")
        without_prefix = getattr(model_cls, "__without_prefix__", ())
        if not prefix or set(names).intersection(without_prefix):
            return (*names, *map(str.upper, names))

        processed_names = []
        for name in names:
            name.lstrip("_")
            processed = f"{prefix}_{name}".lower()
            processed_names.extend((processed, processed.upper()))

        return tuple(processed_names)

    @classmethod
    def resolve_confignames(
        cls,
        root: Path,
    ) -> dict[type, tuple[Sequence[str], ...]]:
        """
        The resolve_confignames function resolves the names of all environment
        variables required by the configs in any class marked as config.
        It returns them in a dict of {class: (configs)}.

        :param cls: Call the class_validator function, which is used to validate that a
        class has all of the required attributes
        :param root: Path: Specify the root directory of the project
        :return: A dictionary of {class: (configs)}
        """
        provider_finder = finder.Finder(is_config, root)
        provider_finder.find()
        tempself = cls()
        return {
            provider: tuple(
                tempself.resolve_names(provider, field, "")
                for field in provider.__fields__.values()
            )
            for provider in provider_finder.output.values()
        }



from .asyncio import AsyncSaAdapter
from .asyncio import AsyncSaContext
from .sync import SaAdapter
from .sync import SaContext

__all__ = ["AsyncSaContext", "AsyncSaAdapter", "SaAdapter", "SaContext"]



from dataclasses import MISSING as dc_MISSING
from dataclasses import Field
from typing import Any
from typing import Generator
from typing import Sequence
from typing import Union

from attrs import define

from gyver.config.adapter.interface import FieldResolverStrategy
from gyver.config.config import MISSING


@define(frozen=True, slots=True)
class DataclassResolverStrategy(FieldResolverStrategy[Field]):
    field: Field

    def cast(self) -> type:
        return self.field.type

    def names(self) -> Sequence[str]:
        return (self.field.name,)

    def init_name(self) -> str:
        return self.field.name

    def default(self) -> Union[Any, type[MISSING]]:
        if self.field.default not in (None, Ellipsis, dc_MISSING):
            return self.field.default
        return (
            MISSING
            if self.field.default_factory in (None, dc_MISSING)
            else self.field.default_factory
        )

    @staticmethod
    def iterfield(config_class: type) -> Generator[Field, Any, Any]:
        yield from config_class.__dataclass_fields__.values()  # type: ignore



import typing
from dataclasses import dataclass

import sqlalchemy as sa

from gyver.database.typedef import ClauseType

from . import _helpers
from . import comp as cp
from . import interface

T = typing.TypeVar("T")

CACHE_MAXLEN = 250


class _BindCache:
    def __init__(self) -> None:
        self._cached: dict[typing.Hashable, interface.Comparison] = {}
        self.maxlen = CACHE_MAXLEN

    def get(self, key: typing.Hashable) -> typing.Optional[interface.Comparison]:
        return self._cached.get(key)

    def set(
        self, key: typing.Hashable, value: interface.Comparison
    ) -> interface.Comparison:
        if len(self._cached) >= CACHE_MAXLEN:
            self._cached.pop(tuple(self._cached)[0])
        self._cached[key] = value
        return value


_cache = _BindCache()


@dataclass(frozen=True)
class Resolver(typing.Generic[T]):
    val: typing.Any

    def resolve(self, mapper: interface.Mapper):
        del mapper
        return self.val

    def __bool__(self):
        return self.val is not None


class FieldResolver(Resolver[str]):
    def resolve(self, mapper: interface.Mapper):
        return _helpers.retrieve_attr(mapper, self.val)


class Where(interface.BindClause, typing.Generic[T]):
    def __init__(
        self,
        field: str,
        expected: typing.Optional[T] = None,
        comp: interface.Comparator[T] = cp.equals,
        resolver_class: type[Resolver[T]] = Resolver,
    ) -> None:
        self.field = field
        self.expected = resolver_class(expected)
        self.comp = comp

    type_ = ClauseType.BIND

    def bind(self, mapper: interface.Mapper) -> interface.Comparison:
        if self.comp is cp.always_true:
            return AlwaysTrue().bind(mapper)
        resolved = self.expected.resolve(mapper)
        if not isinstance(mapper, typing.Hashable):
            return self._get_comparison(
                _helpers.retrieve_attr(mapper, self.field),
                resolved,
            )
        if (value := _cache.get((mapper, self.field, resolved, self.comp))) is not None:
            return value
        attr = _helpers.retrieve_attr(mapper, self.field)

        return _cache.set(
            (mapper, self.field, resolved, self.comp),
            self._get_comparison(attr, resolved),
        )

    def _get_comparison(self, attr: interface.FieldType, resolved: typing.Any):
        return self.comp(attr, resolved) if self.expected else sa.true()


class _JoinBind(interface.BindClause):
    def __init__(
        self,
        items: typing.Sequence[interface.BindClause],
        operator: typing.Callable[..., interface.Comparison],
    ) -> None:
        self.items = items
        self.operator = operator

    type_ = ClauseType.BIND

    def bind(self, mapper: interface.Mapper) -> interface.Comparison:
        return self.operator(*(item.bind(mapper) for item in self.items))


def and_(*bind: interface.BindClause) -> _JoinBind:
    return _JoinBind(bind, sa.and_)


def or_(*bind: interface.BindClause) -> _JoinBind:
    return _JoinBind(bind, sa.or_)


_placeholder_column = sa.Column("placeholder")


class AlwaysTrue(interface.BindClause):
    type_ = ClauseType.BIND

    def bind(self, mapper: interface.Mapper) -> interface.Comparison:
        return cp.always_true(_placeholder_column, mapper)


class RawQuery(interface.BindClause):
    type_ = ClauseType.BIND

    def __init__(self, cmp: interface.Comparison) -> None:
        self._cmp = cmp

    def bind(self, mapper: interface.Mapper) -> interface.Comparison:
        del mapper
        return self._cmp



from typing import Optional

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import padding
from cryptography.hazmat.primitives.asymmetric import rsa

from gyver.config import ConfigLoader
from gyver.config import ProviderConfig
from gyver.utils import lazyfield


class RSACryptoConfig(ProviderConfig):
    __prefix__ = "crypto"

    private_key: str
    public_key: str


class RSACryptoProvider:
    def __init__(self, config: Optional[RSACryptoConfig] = None) -> None:
        self._config = config or ConfigLoader().load(RSACryptoConfig)

    @property
    def config(self):
        return self._config

    @lazyfield
    def _private_key(self) -> rsa.RSAPrivateKey:
        return serialization.load_pem_private_key(  # type: ignore
            self._config.private_key.encode(),
            password=None,
            backend=default_backend(),
        )

    @lazyfield
    def _public_key(self) -> rsa.RSAPublicKey:
        return serialization.load_pem_public_key(  # type:ignore
            self._config.public_key.encode(), backend=default_backend()
        )

    def encrypt(self, secret: str) -> str:
        ciphertext = self._public_key.encrypt(
            secret.encode(),
            padding.OAEP(
                mgf=padding.MGF1(algorithm=hashes.SHA256()),
                algorithm=hashes.SHA256(),
                label=None,
            ),
        )
        return ciphertext.hex()

    def decrypt(self, ciphertext: str) -> str:
        encoded = bytes.fromhex(ciphertext)
        return self._private_key.decrypt(
            encoded,
            padding.OAEP(
                mgf=padding.MGF1(algorithm=hashes.SHA256()),
                algorithm=hashes.SHA256(),
                label=None,
            ),
        ).decode()



import typing

from sanic import Request
from sanic import Sanic

from gyver.database import AsyncSaContext
from gyver.database import SaContext
from gyver.database.config import DatabaseConfig
from gyver.database.provider import AsyncDatabaseProvider
from gyver.database.provider import SyncDatabaseProvider

TrxOptions = typing.Optional[typing.Literal["open", "begin"]]


def setup_database_session(
    app: Sanic,
    config: typing.Optional[DatabaseConfig] = None,
    sync: bool = False,
):
    provider_class = SyncDatabaseProvider if sync else AsyncDatabaseProvider

    async def _setup(_):
        provider = provider_class(config)
        app.ctx.gyver_database = provider

    app.register_listener(_setup, "before_server_start")


def get_async_context(transaction_on: TrxOptions = "open"):
    def _get(request: Request) -> AsyncSaContext:
        provider = typing.cast(AsyncDatabaseProvider, request.app.ctx.gyver_database)
        return provider.context(transaction_on=transaction_on)

    return _get


def get_sync_context(transaction_on: TrxOptions = "open"):
    def _get(request: Request) -> SaContext:
        provider = typing.cast(SyncDatabaseProvider, request.app.ctx.gyver_database)
        return provider.context(transaction_on=transaction_on)

    return _get
